#include "common.h"
#include "Xmalloc.cu"

#ifndef GPC_H
#define GPC_H

#define XMC_DISABLE_SUPERBLOCKS
#define GPC_EPSILON (DBL_EPSILON)

#define GPC_VERSION "2.32"

#ifndef TRUE
#define FALSE              0
#define TRUE               1
#endif

#define LEFT               0
#define RIGHT              1

#define ABOVE              0
#define BELOW              1

#define CLIP               0
#define SUBJ               1

#define INVERT_TRISTRIPS   FALSE
 
#define EQ(a, b)           (fabs((a) - (b)) <= GPC_EPSILON)

#define PREV_INDEX(i, n)   ((i - 1 + n) % n)
#define NEXT_INDEX(i, n)   ((i + 1    ) % n)

#define OPTIMAL(v, i, n)   ((v[PREV_INDEX(i, n)].y != v[i].y) || \
                            (v[NEXT_INDEX(i, n)].y != v[i].y))

#define FWD_MIN(v, i, n)   ((v[PREV_INDEX(i, n)].vertex.y >= v[i].vertex.y) \
                         && (v[NEXT_INDEX(i, n)].vertex.y > v[i].vertex.y))

#define NOT_FMAX(v, i, n)   (v[NEXT_INDEX(i, n)].vertex.y > v[i].vertex.y)

#define REV_MIN(v, i, n)   ((v[PREV_INDEX(i, n)].vertex.y > v[i].vertex.y) \
                         && (v[NEXT_INDEX(i, n)].vertex.y >= v[i].vertex.y))

#define NOT_RMAX(v, i, n)   (v[PREV_INDEX(i, n)].vertex.y > v[i].vertex.y)

#define VERTEX(e,p,s,x,y)  {add_vertex(&((e)->outp[(p)]->v[(s)]), x, y); \
                            (e)->outp[(p)]->active++;}

#define P_EDGE(d,e,p,i,j)  {(d)= (e); \
                            do {(d)= (d)->prev;} while (!(d)->outp[(p)]); \
                            (i)= (d)->bot.x + (d)->dx * ((j)-(d)->bot.y);}

#define N_EDGE(d,e,p,i,j)  {(d)= (e); \
                            do {(d)= (d)->next;} while (!(d)->outp[(p)]); \
                            (i)= (d)->bot.x + (d)->dx * ((j)-(d)->bot.y);}


#define MALLOC(p, b, s, t) {if ((b) > 0) { \
                            p= (t*)xmcMalloc(b); if (!(p)) { \
                            printf("gpc malloc failure: %s\n", s); \
                            }} else p= NULL;}

#define MALLOC_CPU(p, b, s, t) {if ((b) > 0) { \
                            p= (t*)malloc(b); if (!(p)) { \
                            printf("gpc malloc failure: %s\n", s); \
                            }} else p= NULL;}

#define FREE(p)            {if (p) {xmcFree(p); (p)= NULL;}}

#define FREE_CPU(p)            {if (p) {free(p); (p)= NULL;}}

typedef enum                        /* Edge intersection classes         */
{
  NUL,                              /* Empty non-intersection            */
  EMX,                              /* External maximum                  */
  ELI,                              /* External left intermediate        */
  TED,                              /* Top edge                          */
  ERI,                              /* External right intermediate       */
  RED,                              /* Right edge                        */
  IMM,                              /* Internal maximum and minimum      */
  IMN,                              /* Internal minimum                  */
  EMN,                              /* External minimum                  */
  EMM,                              /* External maximum and minimum      */
  LED,                              /* Left edge                         */
  ILI,                              /* Internal left intermediate        */
  BED,                              /* Bottom edge                       */
  IRI,                              /* Internal right intermediate       */
  IMX,                              /* Internal maximum                  */
  FUL                               /* Full non-intersection             */
} vertex_type;

typedef enum                        /* Horizontal edge states            */
{
  NH,                               /* No horizontal edge                */
  BH,                               /* Bottom horizontal edge            */
  TH                                /* Top horizontal edge               */
} h_state;

typedef enum                        /* Edge bundle state                 */
{
  UNBUNDLED,                        /* Isolated edge not within a bundle */
  BUNDLE_HEAD,                      /* Bundle head node                  */
  BUNDLE_TAIL                       /* Passive bundle tail node          */
} bundle_state;

typedef struct                      /* Polygon vertex structure          */
{
  double              x;            /* Vertex x component                */
  double              y;            /* vertex y component                */
} gpc_vertex;

typedef struct                      /* Vertex list structure             */
{
  	int                 num_vertices;
  	gpc_vertex         *vertex;       /* Vertex array pointer              */
} gpc_vertex_list;

typedef struct                      /* Polygon set structure             */
{
  int                 num_contours; /* Number of contours in polygon     */
  int                *hole;         /* Hole / external contour flags     */
  gpc_vertex_list    *contour;      /* Contour array pointer             */
} gpc_polygon;

typedef struct v_shape              /* Internal vertex list datatype     */
{
  double              x;            /* X coordinate component            */
  double              y;            /* Y coordinate component            */
  struct v_shape     *next;         /* Pointer to next vertex in list    */
} vertex_node;

typedef struct p_shape              /* Internal contour / tristrip type  */
{
  int                 active;       /* Active flag / vertex count        */
  int                 hole;         /* Hole / external contour flag      */
  vertex_node        *v[2];         /* Left and right vertex list ptrs   */
  struct p_shape     *next;         /* Pointer to next polygon contour   */
  struct p_shape     *proxy;        /* Pointer to actual structure used  */
} polygon_node;

typedef struct edge_shape
{
  gpc_vertex          vertex;       /* Piggy-backed contour vertex data  */
  gpc_vertex          bot;          /* Edge lower (x, y) coordinate      */
  gpc_vertex          top;          /* Edge upper (x, y) coordinate      */
  double              xb;           /* Scanbeam bottom x coordinate      */
  double              xt;           /* Scanbeam top x coordinate         */
  double              dx;           /* Change in x for a unit y increase */
  int                 type;         /* Clip / subject edge flag          */
  int                 bundle[2][2]; /* Bundle edge flags                 */
  int                 bside[2];     /* Bundle left / right indicators    */
  bundle_state        bstate[2];    /* Edge bundle state                 */
  polygon_node       *outp[2];      /* Output polygon / tristrip pointer */
  struct edge_shape  *prev;         /* Previous edge in the AET          */
  struct edge_shape  *next;         /* Next edge in the AET              */
  struct edge_shape  *pred;         /* Edge connected at the lower end   */
  struct edge_shape  *succ;         /* Edge connected at the upper end   */
  struct edge_shape  *next_bound;   /* Pointer to next bound in LMT      */
} edge_node;

typedef struct lmt_shape            /* Local minima table                */
{
  double              y;            /* Y coordinate at local minimum     */
  edge_node          *first_bound;  /* Pointer to bound list             */
  struct lmt_shape   *next;         /* Pointer to next local minimum     */
} lmt_node;

typedef struct sbt_t_shape          /* Scanbeam tree                     */
{
  double              y;            /* Scanbeam node y value             */
  int visited;
  struct sbt_t_shape *less;         /* Pointer to nodes with lower y     */
  struct sbt_t_shape *more;         /* Pointer to nodes with higher y    */
} sb_tree;

typedef struct it_shape             /* Intersection table                */
{
  edge_node          *ie[2];        /* Intersecting edge (bundle) pair   */
  gpc_vertex          point;        /* Point of intersection             */
  struct it_shape    *next;         /* The next intersection table node  */
} it_node;

typedef struct st_shape             /* Sorted edge table                 */
{
  edge_node          *edge;         /* Pointer to AET edge               */
  double              xb;           /* Scanbeam bottom x coordinate      */
  double              xt;           /* Scanbeam top x coordinate         */
  double              dx;           /* Change in x for a unit y increase */
  struct st_shape    *prev;         /* Previous edge in sorted list      */
} st_node;

typedef struct bbox_shape           /* Contour axis-aligned bounding box */
{
  double             xmin;          /* Minimum x coordinate              */
  double             ymin;          /* Minimum y coordinate              */
  double             xmax;          /* Maximum x coordinate              */
  double             ymax;          /* Maximum y coordinate              */
} bbox;

typedef enum                        /* Set operation type                */
{
  GPC_DIFF,                         /* Difference                        */
  GPC_INT,                          /* Intersection                      */
  GPC_XOR,                          /* Exclusive or                      */
  GPC_UNION                         /* Union                             */
} gpc_op;


__device__ void gpc_polygon_clip(gpc_op op, gpc_polygon *subj, gpc_polygon *clip, gpc_polygon *result);
void printfOutput(gpc_polygon *result,int numOfOutput);
void run(rtree_search * output_after, int numOfNodes,int * num_base,long int * prefix_base,gpc_vertex * vertice_base,int *num_overlay,long int * prefix_overlay, gpc_vertex * vertice_overlay,int numOfOutput,int *hole_base,int *hole_overlay,bounding_box *zOrder);
__global__ void Execute(rtree_search *output,int offset,int * num_base,long int * prefix_base,gpc_vertex * vertice_base,int *,long int * prefix_overlay, gpc_vertex * vertice_overlay,int num,gpc_polygon * result,int *hole_base,int *hole_overlay,bounding_box *zOrder);
__device__ void gpc_free_polygon(gpc_polygon *p);
void readData(char * base_filename,char * overlay_filename, int **rect1,int **rect2,int **rect3,int **rect4,int * num_contours,int **rect1_query,int **rect2_query,int **rect3_query,int **rect4_query,int *numOfQuerys,int ** num_base,long int **prefix_base,gpc_vertex ** vertex_base,int **num_overlay,long int **prefix_overlay,gpc_vertex **vertex_overlay,int *minX,int *minY,int **hole_base_cpu,int **hole_overlay_cpu);

__device__ void reset_it(it_node **it);
__device__  void reset_lmt(lmt_node **lmt);
__device__  void insert_bound(edge_node **b, edge_node *e);
__device__  edge_node **bound_list(lmt_node **lmt, double y);
__device__  void add_to_sbtree(int *entries, sb_tree **sbtree, double y);
__device__ void build_sbt(int *entries,double *sbt,sb_tree  *sbtree);
__device__  void free_sbtree(sb_tree **sbtree);                              
__device__  int count_optimal_vertices(gpc_vertex_list c);
__device__  edge_node *build_lmt(lmt_node **lmt, sb_tree **sbtree,int *sbt_entries, gpc_polygon *p, int type,gpc_op op);
__device__  void add_edge_to_aet(edge_node **aet, edge_node *edge, edge_node *prev);              
__device__  void add_intersection(it_node **it, edge_node *edge0, edge_node *edge1,double x, double y);
__device__  void add_st_edge(st_node **st, it_node **it, edge_node *edge, double dy);
__device__  void build_intersection_table(it_node **it, edge_node *aet, double dy);
__device__  int count_contours(polygon_node *polygon);
__device__  void add_left(polygon_node *p, double x, double y);
__device__  void merge_left(polygon_node *p, polygon_node *q, polygon_node *list);
__device__  void add_right(polygon_node *p, double x, double y);
__device__  void merge_right(polygon_node *p, polygon_node *q, polygon_node *list);
__device__  void add_local_min(polygon_node **p, edge_node *edge,double x, double y);
__device__  bbox *create_contour_bboxes(gpc_polygon *p);
__device__  void minimax_test(gpc_polygon *subj, gpc_polygon *clip,gpc_op op);

#endif